import java.net.Socket;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

class Connection implements Runnable, AgentMonitor
{ 
   private String            mDestHost         = null;
   private int               mDestPort         = -1;
   private Agent             mFromSrcToDest    = null;
   private Agent             mFromDestToSrc    = null;
   private Socket            mSrcSocket        = null;
   private Socket            mDestSocket       = null;
   private InputStream       mSrcIn            = null;
   private OutputStream      mSrcOut           = null;
   private InputStream       mDestIn           = null;
   private OutputStream      mDestOut          = null;
   private ConnectionMonitor mMonitor          = null;
   private boolean           mConnectionClosed = false;
   private boolean           mEcho             = false;
   private boolean           mBlock            = false;
   private boolean           mXBlock           = false;
   private int               mID               = 0;

   public Connection
   (
      Socket s,
      ConnectionMonitor monitor,
      String destHost,
      int destPort,
      boolean echo,
      boolean block,
      boolean xBlock,
      int idNumber
   )
   {
      mSrcSocket = s;
      mMonitor   = monitor;
      mDestHost  = destHost;
      mDestPort  = destPort;
      mEcho      = echo;
      mBlock     = block;
      mXBlock    = xBlock;
      this.mID   = idNumber;

      mMonitor.attemptingConnection(this);

      try
      {
         // Establish read/write for the socket

         mSrcIn  = mSrcSocket.getInputStream();
         mSrcOut = mSrcSocket.getOutputStream();

         // Start ourself, so there's no delay in getting back 
         // to the server to listen for new connections

         Thread t = new Thread(this);
         t.start();
      }
      catch(IOException e)
      {
         mMonitor.connectionError(this, "" + e);
      }
   }

   public void run()
   {
      if(!connectToDest())
      {
         closeSrc();
      }
      else
      {
         // Ok, we're all ready ... since we've gotten this far, 
         // add ourselves into the connection list

         mMonitor.addConnection(this);

         // Create our two agents

         mFromSrcToDest = new Agent(
            mSrcIn,
            mDestOut,
            this,
            "src  => dest",
            mEcho,
            mBlock,
            mXBlock,
            mID
         );

         mFromDestToSrc = new Agent(
            mDestIn,
            mSrcOut,
            this,
            "dest => src",
            mEcho,
            mBlock,
            mXBlock,
            mID
         );

         // No need for our thread to continue, we'll be notified if
         // either of our agents dies
      }
   }

   public synchronized void agentHasDied(Agent a)
   {
      // When one agent dies, so will the other ... if the
      // connection is already closed then we have already been
      // visited by the first agent ... just return

      if(mConnectionClosed) return;

      closeSrc();
      closeDest();

      mMonitor.removeConnection(this);
      mConnectionClosed = true;
   }

   private boolean connectToDest()
   {
      // Ok, we've got the host name and port to which we wish to
      // connect, try to establish a connection

      try
      {
         mDestSocket = new Socket(mDestHost, mDestPort);
         mDestIn     = mDestSocket.getInputStream();
         mDestOut    = mDestSocket.getOutputStream();
      }
      catch(Exception e)
      {
         mMonitor.connectionError(this, "connect error: "
           + mDestHost + "/" + mDestPort + " " + e);
         return(false);
      }

      return(true);
   }

   private void closeSrc()
   {
      try
      {
         mSrcIn.close(); mSrcOut.close(); mSrcSocket.close();
      }
      catch(Exception e) {}
   }

   private void closeDest()
   {
      try
      {
         mDestIn.close(); mDestOut.close(); mDestSocket.close();
      }
      catch(Exception e) {}
   }

   public String getSrcHost()
   {
      return(mSrcSocket.getInetAddress().toString());
   }

   public String getDestHost()
   {
      return(mDestSocket.getInetAddress().toString());
   }

   public int getDestPort()
   {
      return(mDestPort);
   }
}
